-- A. Tasks and task categories replace service items and catalog service items so 
-- 1. create new task and taskcategory tables and 
-- 2. populate them with service items and catalog service items where itemtype :{service = 0, expense = 1, product = 2}
CREATE TABLE taskcategory(
	taskcategoryid BIGINT GENERATED BY DEFAULT AS IDENTITY(START WITH 1) NOT NULL,
	uuid VARCHAR(36) NOT NULL,
	created TIMESTAMP NOT NULL,
	updated TIMESTAMP,
	businessid BIGINT,
	name VARCHAR(15) NOT NULL,
	description VARCHAR(100) NOT NULL,
	enabled BOOLEAN,
	billable BOOLEAN,
	timeroundingmode INTEGER,
	timeroundinginterval INTEGER,
	timeroundingcalculation INTEGER,
	timeroundingminimuminterval INTEGER,
	timeroundingenabled BOOLEAN,
	priceamount NUMERIC,
	pricecurrency VARCHAR(255),											
	quantitytype INTEGER,
	unitofmeasure VARCHAR(50),
	CONSTRAINT fk_taskcategory_businessid FOREIGN KEY (businessid) REFERENCES PUBLIC.business(businessid)
);

CREATE TABLE task (
	taskid BIGINT GENERATED BY DEFAULT AS IDENTITY(START WITH 1) NOT NULL,
	uuid VARCHAR(36) NOT NULL,
	created TIMESTAMP NOT NULL,
	updated TIMESTAMP,
	projectid BIGINT,
	name VARCHAR(15) NOT NULL,
	description VARCHAR(100),
	tags VARCHAR(255),
	notes VARCHAR(250),
	startdate TIMESTAMP,
	duedate TIMESTAMP,
	completeddate TIMESTAMP,
	completed BOOLEAN,
	estimatedtime BIGINT,
	billable BOOLEAN,
	taskcategoryid BIGINT,
	taxexempt BOOLEAN,
	timeroundingmode INTEGER,
	timeroundinginterval INTEGER,
	timeroundingcalculation INTEGER,
	timeroundingminimuminterval INTEGER,
	timeroundingenabled BOOLEAN,
	billableunits NUMERIC,
	billablequantitydiscountenabled BOOLEAN,
	billablequantitydiscounttype INTEGER,
	billablequantitydiscountpercentagevalue NUMERIC,
	billablequantitydiscountfixedvalue NUMERIC,
	quantitytype INTEGER,
	unitofmeasure VARCHAR(50),
	pricediscountenabled BOOLEAN,
	pricediscounttype INTEGER,
	pricediscountpercentagevalue NUMERIC,
	pricediscountamount NUMERIC,
	pricediscountcurrency VARCHAR(255),
	priceamount NUMERIC,
	pricecurrency VARCHAR(255),											
	PRIMARY KEY(taskid),
	CONSTRAINT fk_task_projectid FOREIGN KEY (projectid) REFERENCES PUBLIC.project(projectid),
	CONSTRAINT fk_task_taskcategoryid FOREIGN KEY (taskcategoryid) REFERENCES PUBLIC.taskcategory(taskcategoryid)
);	

INSERT INTO taskcategory (taskcategoryid,uuid,created,updated,businessid,name,enabled,description,billable,unitofmeasure,priceamount,pricecurrency,quantitytype,timeroundingmode,timeroundinginterval,timeroundingcalculation,timeroundingminimuminterval,timeroundingenabled) 
SELECT catalogitemid,uuid,created,updated,businessid,name,enabled,description,true,unitofmeasure,rate,currency,quantitytype,1,900,0,0,FALSE
FROM catalogitem
WHERE catalogitem.itemtype=0;

INSERT INTO task (taskid,uuid,created,updated,projectid,name,description,notes,tags,startdate,taskcategoryid,duedate,completeddate,completed,estimatedtime,billable,taxexempt,quantitytype,unitofmeasure,timeroundingmode,timeroundinginterval,timeroundingcalculation,timeroundingminimuminterval,timeroundingenabled,billableunits,billablequantitydiscountenabled,billablequantitydiscounttype,billablequantitydiscountpercentagevalue,billablequantitydiscountfixedvalue,pricediscountenabled,pricediscounttype,pricediscountpercentagevalue,pricediscountamount,pricediscountcurrency,priceamount,pricecurrency) 
SELECT itemid,uuid,created,updated,projectid,name,description,notes,'',date,catalogitemid,null,null,false,0,billable,taxexempt,quantitytype,unitofmeasure,roundingmode,roundinginterval,roundingcalculation,minimuminterval,enabled,units,unitsdiscountenabled,unitsdiscounttype,unitsdiscountpercentagevalue,unitsdiscountfixedvalue,ratediscountenabled,ratediscounttype,ratediscountpercentagevalue,ratediscountrate,ratediscountcurrency,rate,currency
FROM item
WHERE item.itemtype=0;

-- B. Clean item and catalogitem tables once tasks and task categories were imported so
-- 1. set catalog item to null for service items because service items will no longer use catalog items
-- 2. delete service catalog items because they were replaced by task categories and service items don't use then anymore
-- 3. drop unused catalog item and item columns
UPDATE item SET catalogitemid=null WHERE itemtype=0;
DELETE FROM catalogitem WHERE itemtype=0;

ALTER TABLE catalogitem DROP COLUMN itemtype;
ALTER TABLE catalogitem DROP COLUMN quantitytype;
ALTER TABLE item DROP COLUMN billable;
ALTER TABLE item DROP COLUMN date;
ALTER TABLE item DROP COLUMN notes;

-- C. Time entries are saved to tasks and billed by service items so we have to 
-- 1. link all time entries to the corresponding tasks and
-- 2. unlink time entries from not invoiced service items because these items will be deleted.
ALTER TABLE timeentry ADD taskid BIGINT;
ALTER TABLE timeentry ADD CONSTRAINT fk_timeentry_taskid FOREIGN KEY (taskid) REFERENCES PUBLIC.task(taskid);

UPDATE timeentry SET taskid=itemid;
UPDATE timeentry SET itemid=null WHERE EXISTS (SELECT itemid FROM item WHERE timeentry.itemid=item.itemid AND item.invoiceid is null AND item.itemtype=0);

-- D. A service item bills time from a task and it is linked with the task (so it can calculate its billed total) and with the invoice so
-- 1. remove all not invoiced service items because they were replaced by tasks and their time is now saved on tasks and
-- 2. link invoiced service items to their task and 
-- 3. remove service items from projects because they only belong to an invoice now.
ALTER TABLE item ADD taskid BIGINT;
ALTER TABLE item ADD CONSTRAINT fk_item_taskid FOREIGN KEY (taskid) REFERENCES PUBLIC.task(taskid);

DELETE FROM item WHERE invoiceid is null AND itemtype=0;
UPDATE item SET taskid=itemid WHERE invoiceid is not null AND itemtype=0;
UPDATE item SET projectid=null WHERE invoiceid is not null AND itemtype=0;

-- E. Invoiced tasks must be marked as completed because that was the behavior of service items
-- 1. mark invoiced tasks as completed and reset their billable units to 0 (their quantity has already been invoiced so there is nothing else to bill)
-- 2. the completed date of a task is the date of the invoice that bills it
UPDATE task SET completed=true,billableunits=0 WHERE EXISTS (SELECT taskid FROM item WHERE task.taskid=item.taskid);
UPDATE task SET completeddate=(SELECT invoice.date FROM invoice JOIN item ON item.invoiceid=invoice.invoiceid WHERE item.taskid=task.taskid);

-- F. Project settings for services are now settings for tasks
ALTER TABLE project ALTER COLUMN serviceBillable RENAME TO taskBillable;
ALTER TABLE project ALTER COLUMN serviceQuantityType RENAME TO taskQuantityType;
ALTER TABLE project ALTER COLUMN servicePrice RENAME TO taskPrice;
ALTER TABLE project ALTER COLUMN servicePriceCurrency RENAME TO taskPriceCurrency;
ALTER TABLE project ALTER COLUMN serviceUnitOfMeasure RENAME TO taskUnitOfMeasure;
ALTER TABLE project ALTER COLUMN serviceTimeRoundingMode RENAME TO taskTimeRoundingMode;
ALTER TABLE project ALTER COLUMN serviceTimeRoundingCalculation RENAME TO taskTimeRoundingCalculation;
ALTER TABLE project ALTER COLUMN serviceTimeRoundingInterval RENAME TO taskTimeRoundingInterval;
ALTER TABLE project ALTER COLUMN serviceTimeMinimumInterval RENAME TO taskTimeMinimumInterval;
ALTER TABLE project ALTER COLUMN serviceTimeRoundingEnabled RENAME TO taskTimeRoundingEnabled;
